<?php
class Lexan {

	// konstanty které vrací lexikální analyzátor jako prvky jazyka
	const SYMB_EOI = "EOI";
	const SYMB_INT = "INT";
	const SYMB_REAL = "FLOAT";
	const SYMB_HEXA = "HEXA";
	const SYMB_BOOL = "BOOL";
	const SYMB_VAR = "VAR";
	const SYMB_IDENT = "IDENT";
	const SYMB_COMMENT = "COMMENT";
	const SYMB_STRING = "STRING";
	const SYMB_WHITESPACE = "WHITESPACE";
	const SYMB_HTML = "HTML";
	const DOLLAR_SYMB = "DOLLAR";
	
	// flagy
	const _ADD_LINE_NUMBERS = '001';

	private $kw; // objekt typu KeyWord
	private $in; // vstup (zadaný zdrojový kód)
	private $out = ""; // výstup - zvýrazněné HTML
	private $syntax = ""; // výstup - posloupnost lexikálních elementů pro syntaktický analyzátor. Syntaktickému analyzátoru se nepředávají WHITESPACE a COMMENT
	private $symbols = array(); // seznam symbolů ze $syntax, kde su uchovávají informace o obsahu a umístění lexikálního elementu
	private $line = 1; // počet řádků programu
	private $inHtml = true; // zda se analyzátor nachází v HTML části kódu
	
	private $position = 0; // aktuální pozice ve vstupním kódu
	private $char = ''; // aktuální znak který lexikální analyzátor kontroluje
	
	/**
	* Konstruktor, inicializuje analyzátor do základního stavu.
	* 
	* @access public 
	* @param string $string - zdrojový kód který chceme analyzovat
	* @param KeyWord $kw - abeceda jazyka
	*/
	public function __construct($string, KeyWord &$kw)
	{
		$this->in = $string;
		$this->kw = $kw;
	}
	
	/**
	* Spustí lexikální analýzu kódu předného konstruktoru
	* 
	* @access public
	* @param int $flags - modifikátory (v současné době jen na zapnutí/vypnutí řádkování)
	*/
	public function analyze($flags = 0)
	{
		$showLineNumbers = ($flags & self::_ADD_LINE_NUMBERS);
		$this->readChar();
		list($symb, $value, $from, $to) = $this->readSymb();
		if ($showLineNumbers && $symb !== self::SYMB_EOI) {
			$this->out = "<span id=\"line_".$from."\" class=\"lineNumber\">".($from)."\t</span>";	
		}
		while ($symb !== self::SYMB_EOI) {
			if ($symb != self::SYMB_COMMENT && $symb != self::SYMB_WHITESPACE) {
				$this->symbols [] = array ( $symb, $value, $from, $to );
				$parts = explode(" ", $symb);
				if (isset($parts[1]) && $parts[0] === trim(KeyWord::KEYWORD_PREFIX)) {
					$this->syntax .= strtoupper(str_replace(" ", "_",$symb)) . " ";
				} else if (isset($parts[1])) {
					$this->syntax .= strtoupper($parts[1]) . " ";
				} else {
					$this->syntax .= strtoupper(str_replace(" ", "_",$symb)) . " ";
				}
			}
			if($from == $to) {
				$out = "<span class=\"line_".($from)."\">";
			} else {
				$out = "";
			}
			while($showLineNumbers && $place = strpos($value, "\n")) {
				$lineBreak = true;
				$out .= htmlspecialchars(substr($value, 0, $place + 1))."<span id=\"l".($from + 1)."\" class=\"lineNumber ".(($from + 1) % 10 == 0 ? "bold" : "")."\">".(++$from)."\t</span>";
				$value = substr($value, $place + 1);
			}
			$out .= htmlspecialchars($value). "</span>";
			$this->out .= "<span class=\"$symb\">".$out."</span>";
			list($symb, $value, $from, $to) = $this->readSymb();
		}
		$this->out .= "</span>";
		$this->syntax .= self::SYMB_EOI;
	}
	
	/**
	* Vypíše (pomocí print_r) do bloku <xmp> pole symbolů které byly zatím 
	* uloženy do syntaktického výstupu.
	* V podstatě použitelné pouze pro debugging. 
	* 
	* @access public 
	*/
	public function printResults(){
		echo "<xmp>".print_r($this->symbols, true)."</xmp>";
	}
	
	/**
	* Vypíše syntaktický výstup.
	* 
	* @access public
	*/
	public function printSyntax() {
		echo $this->syntax;
	}
	
	public function getSyntax() {
		return $this->syntax;
	}
	
	public function getOut(){
		return $this->out;
	}

	/**
	* Zjistí který lexikální element je právě na vstupu ze zdrojového kódu.
	* 
	* @access public
	* @return array($symbol, $value, $lineStart, $lineEnd)
	*/
	private function readSymb()
	{
		$symb = false;
		$value = "";
		$lineStart = $this->line;
		$startPosition = $this->position;
		if ( $this->inHtml ) {
			// pokud jsme právě v HTML čekáme na otevírací tag php(<? nebo <%)
			$symb = self::SYMB_HTML;
			while( $this->char !== false ){
				if ($this->char === '<') {
					if ($this->lookAtNext() === "?" || $this->lookAtNext() === "%") {
						break;
					}
				}
				$value .= $this->char;
				$this->readChar();
			}
			$this->inHtml = false;
		 } else if (ctype_space($this->char)) {
			// odbourání bílých znaků
			while( ctype_space($this->char) ) {
				$symb = self::SYMB_WHITESPACE;
				$value .= $this->char;
				$this->readChar();
			}
		}
		else if ($this->char === false) {
			// konec souboru
			$symb = self::SYMB_EOI;
		} else if ($this->char == '$') {
			// proměnná jazyka PHP
			$symb = self::SYMB_VAR;
			$value .= $this->char;
			$this->readChar();
			while($this->char == '$'){
				// "ukazatel" na PHP proměnnou
				$value .= $this->char;
				$this->readChar();
			}
			if (ctype_alpha($this->char) || $this->char === '_'){
				// název proměnné
				$value .= $this->char;
				$this->readChar();
				while(ctype_alnum($this->char) || $this->char === '_'){
					$value .= $this->char;
					$this->readChar();
				}
			} else if ($this->char === "{" || ctype_space($this->char)) {
				// proměnná typu $ { $etwas }
				$symb = self::DOLLAR_SYMB;
			} else {
				// nesmysl (třeba $4, $-kk apod)
				throw new LexException(LexException::EXPECTED_ALPHA, mb_substr($this->in, $startPosition-1, $this->position - ($startPosition - 1)), ($lineStart . " - " . $this->line));
			}
		} else if (ctype_digit($this->char)) {
			// číslo (INT, FLOAT, HEXA)
			$value .= $this->char;
			if ($this->char === "0" && ($this->lookAtNext() === "x" || $this->lookAtNext() === "X")) {
				// hexadecimální číslo ve tvaru 0x0000
				$this->readChar();
				$value .= $this->char;
				$this->readChar();
				if(!ctype_xdigit($this->char)) {
					// chyba - znak není hexadecimální číslice
					throw new LexException(LexException::EXPECTED_HEXA, mb_substr($this->in, $startPosition-1, $this->position - ($startPosition - 1)), ($lineStart . " - " . $this->line));
				}
				$symb = self::SYMB_HEXA;
				while(ctype_xdigit($this->char)){
					$value .= $this->char;
					$this->readChar();
				}
			}
			if (!$symb) {
				// pokud znak není hexa
				$symb = self::SYMB_INT;
				$this->readChar();
				while(ctype_digit($this->char)){
					$value .= $this->char;
					$this->readChar();
				}
				$value = (int)$value;
				if ($this->char === '.') {
					// narazili jsme na desetinnou čárku, jde o FLOAT
					$symb = self::SYMB_REAL;
					$value .= $this->char;
					$this->readChar();
					while(ctype_digit($this->char)){
						$value .= $this->char;
						$this->readChar();
					}
					$value = (float)$value;
				}
			}
		} else if ($this->char === "." && ctype_digit($this->lookAtNext())) {
			// desetinné číslo začínající .
			$symb = self::SYMB_REAL;
			$value .= $this->char;
			$this->readChar();
			while(ctype_digit($this->char)){
				$value .= $this->char;
				$this->readChar();
			}
			$value = (float)$value;
		} else if (ctype_alpha($this->char) || $this->char === '_') {
			// konstanta (nebo identifikátor funkce)
			$value .= $this->char;
			$this->readChar();
			while(ctype_alnum($this->char) || $this->char === '_') {
				$value .= $this->char;
				$this->readChar();
			}
			if (!$symb = $this->kw->isKeyword($value)) {
				// zjištění zda jde o klíčové slovo
				$symb = self::SYMB_IDENT;
			}
		} else if ($this->char === '?' || $this->char === '%') {
			// možná uzavírající tag PHP
			$value .= $this->char;
			$buffer = $this->char;
			$this->readChar();
			if ($this->char !== '>') {
				// není uzavírající, zjistím zda je znak (? nebo %) ve speciálních symbolech
				$symb = $this->kw->isSpecial($buffer);
			} else {
				// je uzavírajícím tagem
				$value .= $this->char;
				$symb = $this->kw->isSpecial($value);
				$this->inHtml = true;
				$this->readChar();
			}
		} else if ($this->char === '"' || $this->char === '\'') {
			// řetězec
			$symb = self::SYMB_STRING;
			$value .= $this->char;
			$open = $this->char;
			$this->readChar();
			while ($this->char !== $open) {
				// čekám na zavírající znak řetězce
				$value .= $this->char;
				if ( $this->char === "\\" ) {
					// přeskočení oescapovaných znaků
					$this->readChar();
					$value .= $this->char;
				} else if ( $this->char === false) {
					// chyba - konec souboru uvnitř řetězce
					throw new LexException(LexException::TERMINATED_STRING, mb_substr($this->in, $startPosition-1, $this->position - ($startPosition - 1)),  ($lineStart . " - " . $this->line));
				}
				$this->readChar();
			}
			$value .= $this->char;
			$this->readChar();
		} else if ($this->char === '/') {
			// možná koentář, jinak speciální symbol
			$value .= $this->char;
			$symb = $this->kw->isSpecial($this->char);
			$this->readChar();
			if ($this->char === '/') {
				// jednořádkový komentář
				$symb = self::SYMB_COMMENT;
				$value .= $this->char;
				$this->readChar();
				while($this->char != "\n" && $this->char !== false){
					if ($this->char == "?" || $this->char == "%") {
						if ($this->lookAtNext() == ">") {
							break;
						}
					}
					$value .= $this->char;
					$this->readChar();
				}
				if ($this->char === "\n" && $this->char !== false) {
					$value .= $this->char;
					$this->readChar();
				}
			} else if ($this->char == "*") {
				// víceřádkový komentář
				while(true){
					$symb = self::SYMB_COMMENT;
					if ($this->char === false) {
						throw new LexException(LexException::UNTERMINATED_COMMENT, mb_substr($this->in, $startPosition-1, $this->position - ($startPosition - 1)),  ($lineStart . " - " . $this->line));
					}
					if ($this->char == "*") {
						$value .= $this->char;
						$this->readChar();
						if ($this->char == "/") {
							break;
						}
					}
					$value .= $this->char;
					$this->readChar();
				}
				$value .= $this->char;
				$this->readChar();
			}
		} else {
			// hledám zde jde o speciální znak - podívám se o tolik znaků dopředu, kolik je 
			// délka nejdelšího speciálního symbolu a zpětně kontroluji zda jde o spec. symbol
			$forwarding = min(KeyWord::$specialMaxLength, strlen($this->in) - $this->position + 1);
			while(($char = $this->lookAtNext($forwarding, true)) !== false && !ctype_space($char) && $forwarding-- && $symb === false){
				$name = $this->kw->isSpecial($char);
				if ($name !== false) {
					$value = $char;
					$symb = $name;
					$this->position += $forwarding;
					$this->readChar();
				}
			}
			if (!$symb) {
				// pokud stále nebyl symbol nalezen, není kód jazykem PHP 
				throw new LexException(LexException::UNKNOWN_SYMBOL, mb_substr($this->in, $startPosition-1, $this->position - ($startPosition - 1)),  ($lineStart . " - " . $this->line));
			}
		}
		$lineEnd = $this->line;
		return array($symb, $value, $lineStart, $lineEnd);
	}
	
	/**
	* Načte následující znak kódu. Zajišťuje počítání řádků.
	* 
	* @access private
	* 
	*/
	private function readChar(){
		if ($this->position < strlen($this->in)) {
			$this->char = $this->in[$this->position++];
		} else {
			$this->char = false;
		}
		if($this->char === "\n") {
			$this->line++;
		}
		//var_dump($this->char);
	}
	
	/**
	* Podívá se na následujících $count znaků a podle $all rozhodne zda je vrátí
	* všechny nebo jen ten poslední.
	* 
	* @access private
	* @param int $count počet znaků o který se má fce podívat dopředu
	* @param boolean $all má vrátit jen poslední znak nebo všechny
	* @return string
	*/
	private function lookAtNext($count = 1, $all = false) {
		if ($this->position + $count-1 < strlen($this->in)) {
			if (!$all) {
				return $this->in[$this->position + $count-1];	
			} else {
				return substr($this->in, $this->position - 1, $count);
			}
		} else if ($all) {
			return substr($this->in, $this->position - 1);
		} else {
			return false;
		}
	}
	
	public function getTerms() {
		return $this->syntax;
	}
	
	/**
	* Vrací informace o daném symbolu
	* 
	* @params int $place
	* @return array($symb, $value, $from, $to)
	*/
	public function getSymbol($place) {
		if ($place >= 0 && $place < sizeof($this->symbols)) {
			return $this->symbols[$place];
		} else {
			return false;
		}
	}
	
	public function getPrintableSymbol($place) {
		if (!list($symb, $value, $from, $to) = $this->getSymbol($place)) {
			return false;
		} else {
			return "$symb($value) -> $from " . ($to > $from ? " - $to" : ""); 
		} 
	}
}
?>